from ply import yacc

from terrapin.lexer import Lexer
from terrapin.exceptions import TemplateError


class Parser(object):

    def __init__(self):

        self.lexer = Lexer()
        self.tokens = self.lexer.tokens
        self.parser = yacc.yacc(module=self)

    def parse(self, template, context):

        self.context = context

        if template:
            return self.parser.parse(template, lexer=self.lexer.lexer)

    def show_tokens(self, template):
        return self.lexer.tokenize(template)

    def p_output_list_1(self, p):
        """output_list : output
        """
        p[0] = ''.join([wrd for wrd in p[1:]])

    def p_output_list_2(self, p):
        """output_list : output_list output
        """
        p[0] = ''.join([wrd for wrd in p[1:]])

    def p_output(self, p):
        """output : block
                  | WS
                  | STRING
                  | WORD
                  | QUOTEDSTRING
                  | variable
        """
        p[0] = ''.join([wrd for wrd in p[1:]])

    def p_variable(self, p):
        'variable : LVARDELIM WORD RVARDELIM'
        p[0] = self.context.get(p[2], '')

    def p_if_else_statement(self, p):
        """block : if_result output_list else output_list end_if
        """
        p[0] = p[2] if p[1] else p[4]

    def p_empty_if_else_statement(self, p):
        """block : if_result else output_list end_if
        """
        p[0] = '' if p[1] else p[3]

    def p_if_statement(self, p):
        """block : if_result output_list end_if
        """
        p[0] = p[2] if p[1] else ''

    def p_truthy_if(self, p):
        """if_result : LCODEDELIM WS IF WS WORD WS RCODEDELIM
        """
        p[0] = True if self.context.get(p[5]) else False

    def p_equality_if(self, p):
        """if_result : LCODEDELIM WS IF WS WORD WS EQ WS QUOTEDSTRING WS RCODEDELIM
        """
        unquoted_string = p[9][1:-1]
        p[0] = True if self.context.get(p[5]) == unquoted_string else False

    def p_non_equality_if(self, p):
        """if_result : LCODEDELIM WS IF WS WORD WS NE WS QUOTEDSTRING WS RCODEDELIM
        """
        unquoted_string = p[9][1:-1]
        p[0] = True if not self.context.get(p[5]) == unquoted_string else False

    def p_else(self, p):
        """else : LCODEDELIM WS ELSE WS RCODEDELIM
        """
        p[0] = None

    def p_end_if(self, p):
        """end_if : LCODEDELIM WS ENDIF WS RCODEDELIM
        """
        p[0] = None

    def p_error(self, p):
        if p:
            raise TemplateError(p.lineno, p.lexpos, p.value)
        else:
            raise TemplateError(0, 0, '')\



import re

from terrapin.lexer import word_regex


def render_simple(template, context):
    """ Render a template by replacing variables """

    return replace_variables(template, context)


def replace_variables(string, variables):
    """ Format a templated value """

    found_variables = find_variables(string)
    for variable in found_variables:
        value = str(variables.get(variable, ''))
        string = string.replace('{{' + variable + '}}', value)
    return string


def find_variables(template):
    """ Find all the used variables

    This is any string that looks like {{var}}
    """

    capture_variable_regex = r'\{\{(' + word_regex + r')\}\}'
    found_variables = re.findall(capture_variable_regex, template)
    return set(found_variables)



from terrapin.parser import Parser
from terrapin.render_simple import find_variables


parser = Parser()


class ObservingContext(dict):
    """ Keep a record of which keys have tried to be accessed """

    def __init__(self, *args, **kwargs):

        self.keys_seen = set()

    def get(self, key, *args, **kwargs):

        self.keys_seen.add(key)
        return super(ObservingContext, self).get(key, *args, **kwargs)


def used_variables(template):
    """ Return a set of used variables in the template """

    if '{%' in template:
        context = ObservingContext()
        parser.parse(template, context)
        return context.keys_seen
    else:
        return find_variables(template)



from terrapin.render_full import render_full
from terrapin.render_simple import render_simple


def render(template, context):
    """ Choose which rendering strategy to use and execute """

    if '{%' in template:
        return render_full(template, context)
    else:
        return render_simple(template, context)




class TemplateError(Exception):
    """Exception raised during parsing of a template

    Attributes:
        line_number: line number of the error
        position: position of the error
        value: value when the error occured
    """

    def __init__(self, line_number, position, value, message=None):

        self.line_number = line_number
        self.position = position
        self.value = value

        if not message:
            if line_number != 0:
                error_mask = "Template error at line {ln}, position {p} at: {v}"
                message = error_mask.format(ln=line_number,
                                            p=position,
                                            v=value)
            else:
                message = "Unknown template error. Have you closed all open tags? e.g. {% endif %}"

        super(TemplateError, self).__init__(message)



from setuptools import setup, find_packages

setup(name='terrapin',
      version="0.0.16",
      packages=find_packages(),
      description='A very lightweight template language',
      url='https://github.com/DistilledLtd/terrapin',
      author='Distilled Ltd',
      author_email='randd@distilled.net',
      install_requires=['ply']
      )



from ply import lex

from terrapin.exceptions import TemplateError

word_regex = r'[A-Za-z0-9_\.]+'


class Lexer(object):

    reserved = (
        'IF', 'ELSE', 'ENDIF'
    )

    tokens = reserved + (
        # Code delimiters
        'LCODEDELIM', 'RCODEDELIM', 'LVARDELIM', 'RVARDELIM',
        # Operators
        'EQ', 'NE',
        # Literals
        'WS', 'WORD', 'QUOTEDSTRING', 'STRING',
    )

    # t_ignore = r' '  # string.whitespace

    def token_type(self, value, default):

        if value.upper() in self.reserved:
            return value.upper()
        else:
            return default

    def __init__(self):
        self.lexer = lex.lex(module=self)

    def tokenize(self, template):
        """A debug method to print out the tokens generated by the lexer"""
        self.lexer.input(template)
        while True:
            tok = self.lexer.token()
            if tok:
                yield tok
            else:
                break

    def t_LCODEDELIM(self, t):
        r'{%'
        return t

    def t_RCODEDELIM(self, t):
        r'%}'
        return t

    def t_LVARDELIM(self, t):
        r'{{'
        return t

    def t_RVARDELIM(self, t):
        r'}}'
        return t

    def t_EQ(self, t):
        r'=='
        return t

    def t_NE(self, t):
        r'!='
        return t

    @lex.TOKEN(word_regex)
    def t_WORD(self, t):

        t.type = self.token_type(t.value, "WORD")
        return t

    def t_WS(self, t):
        r'[\s]+'
        t.lexer.lineno += t.value.count("\n")
        return t

    def t_QUOTEDSTRING(self, t):
        r"""[\"].*?[\"]|[\'].*?[\']"""
        return t

    def t_STRING(self, t):
        r'([^\{\}]|[\{](?!%|{)|(?<!%|})[\}])+'
        t.type = self.token_type(t.value, "STRING")
        return t

    def t_error(self, t):
        raise TemplateError(t.lexer.lineno, t.lexer.lexpos, t.value[:10])

if __name__ == '__main__':

    s = """{% if in_context %}{% if in_context == "Working" %}Working ^^ hello{% endif %}{% endif %}"""
    toks = Lexer().tokenize(s)
    for tok in toks:
        print(tok)
